library ieee;
use ieee.std_logic_1164.all;
use work.router_pkg.all;

entity fifo is
	generic(
		N : integer := 2
	);
	port(
		in_clk             : in  std_logic;
		in_reset           : in  std_logic;
		in_router_port_p1  : in  router_port_p1_t;
		in_router_port_p2  : out router_port_p2_t;
		out_router_port_p1 : out router_port_p1_t;
		out_router_port_p2 : in  router_port_p2_t
	);
end entity fifo;

architecture RTL of fifo is
	constant SIZE : integer := 2 ** N;
	type data_t is record
		address : router_address_t;
		data    : router_data_t;
	end record;
	type data_array_t is array (0 to SIZE - 1) of data_t;

	type register_t is record
		head : integer range 0 to SIZE - 1;
		tail : integer range 0 to SIZE - 1;
		full : std_logic;
		data : data_array_t;
	end record;

	signal register_reg, register_next : register_t;
	signal router_port_p2              : router_port_p2_t;
	signal router_port_p1              : router_port_p1_t;
begin
	clock : process(in_clk, in_reset) is
	begin
		if (in_reset = '1') then
			register_reg.head <= 0;
			register_reg.tail <= 0;
			register_reg.full <= '0';
			for i in register_reg.data'range loop
				register_reg.data(i).address <= (others => '0');
				register_reg.data(i).data    <= (others => '0');
			end loop;
		elsif (rising_edge(in_clk)) then
			register_reg <= register_next;
		end if;
	end process clock;

	comb : process(in_router_port_p1, out_router_port_p2, register_reg) is
		variable read       : boolean;
		variable tail, head : integer;
	begin
		register_next          <= register_reg;
		router_port_p2.ack     <= '0';
		router_port_p1.address <= register_reg.data(register_reg.head).address;
		router_port_p1.data    <= register_reg.data(register_reg.head).data;
		router_port_p1.write   <= '0';
		read                   := false;
		
		if ((register_reg.head /= register_reg.tail) or (register_reg.full = '1')) then
			router_port_p1.write <= '1';
			if (out_router_port_p2.ack = '1') then
				head := register_reg.head + 1;
				if (head >= size) then
					head := 0;
				end if;
				register_next.head <= head;
				read               := true;
				if (in_router_port_p1.write /= '1') then
					register_next.full <= '0';
				end if;
			end if;
		end if;

		if (in_router_port_p1.write = '1' and (register_reg.full = '0' or read)) then
			register_next.data(register_reg.tail).address <= in_router_port_p1.address;
			register_next.data(register_reg.tail).data    <= in_router_port_p1.data;
			tail                                          := register_reg.tail + 1;
			if (tail >= SIZE) then
				tail := 0;
			end if;
			register_next.tail <= tail;
			router_port_p2.ack <= '1';
			if (register_reg.head = tail and not read) then
				register_next.full <= '1';
			end if;
		end if;
		
	end process comb;

	in_router_port_p2  <= router_port_p2;
	out_router_port_p1 <= router_port_p1;

end architecture RTL;
